/* babtest.c */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

struct Block;
#define BABITED_OBJECT struct Block
#define BABIT_KEY      int
#include "babit.h"

typedef struct Block {
    int value;
    BabitConnector bc;
    int extra;
} Block;

static int BlockCmp (const Block *l, const Block *r)
{
    if (l->value > r->value) return 1;
    else if (l->value < r->value) return -1;
    else return 0;
}

static int BlockChk (const BABIT_KEY *key, const Block *b)
{
    int thisk, thatk;
    int rc;

    thisk = *key;
    thatk = b->value;
    if (thisk > thatk) {
        rc = 1 + (thisk/10 - thatk/10);         /* Greater than  */
    } else if (thisk < thatk) {
        rc = -1 + (thisk/10 - thatk/10);        /* Less than */
    } else {
        rc = 0;					/* Equal to */
    }
/*  fprintf (stderr, "%d ? %d = %d\n", thisk, thatk, rc); */
    return rc;
}

static const BabitControlBlock bcb = {
    0,
    offsetof (Block, bc),
    BlockCmp,
    BlockChk
};

static void FindExTest (BABIT_ROOT *rootp, int key);
static void NextTest   (const BABIT_ROOT *root);
static void WalkTest   (const BABIT_ROOT *root, int option);
static void DeleteTest (BABIT_ROOT *root, int option);
static void InsertTest (BABIT_ROOT *root, int *vect, int vectn);
static void GenSeq     (int *vect, int n, int method);

static int N;
static int debug;
static int check=1;

int main (int argc, char *argv[])
{
    BABIT_ROOT root;
    int *vect;

    debug = 0;
    while (--argc && **++argv == '-') {
        switch (argv[0][1]) {
        case 'd': case 'D':
            debug = 1;
            break;
	case 'c': case 'C':
	    check = 0;
	    break;
        case 0: goto OPTEND;
        default:
            fprintf (stderr, "Unknown option '%s'\n", argv[0]);
            exit (12);
        }
    }
OPTEND:

    if (argc==0) N = 12;
    else {
        N = atoi (argv[0]);
        if (N < 4) {
            fprintf (stderr, "invalid count '%s'\n", argv[0]);
            exit (8);
        }
    }

    srand ((int) time (NULL));
    BabitInit (&root);

    vect = (int *)calloc (N, sizeof (int *));
    if (vect == NULL) {
        fprintf (stderr, "*** Out of memory\n");
        exit (12);
    }
    GenSeq (vect, N, 0);  /* random */
    InsertTest (&root, vect, N);

    NextTest (&root);
    WalkTest (&root, 1+2+4+8);
    FindExTest (&root, N);

    DeleteTest (&root, BABIT_GET_MIN);
/*    FindExTest (&root, N); */

    GenSeq (vect, N, 1); /* novekvo */
    InsertTest (&root, vect, N);
    DeleteTest (&root, BABIT_GET_MED);

    GenSeq (vect, N, -1); /* csokkeno */
    InsertTest (&root, vect, N);
    DeleteTest (&root, BABIT_GET_MAX);

    return 0;
}

static void NextTest (const BABIT_ROOT *root)
{
    Block *obj;
    int dom, he;

    if (check) BabitCheck (root, &dom, &he, &bcb);
/*	i = MAX;
	obj = BabitFind (root, &i, &bcb);*/
	obj = BabitGetMin (root, BABIT_GET_MAX, &bcb);
	while (obj) {
	    printf ("%3d ", obj->value);
	    obj = BabitGetNext (obj, BABIT_GET_PREV, &bcb);
	}
	printf ("\n");

/*	i = MIN;
	obj = BabitFind (root, &i, &bcb);*/
	obj = BabitGetMin (root, BABIT_GET_MIN, &bcb);
	while (obj) {
	    printf ("%3d ", obj->value);
	    obj = BabitGetNext (obj, BABIT_GET_NEXT, &bcb);
	}
	printf ("\n");
}


#define SEPARATOR -99

static struct {
	int option, value;
} TestCodes [] = {
        {BABIT_FIND_LT, 4},
        {BABIT_FIND_EQ, 3},  {BABIT_FIND_LT, 3},
	{BABIT_FIND_EQ, 2},  {BABIT_FIND_LT, 2},
	{BABIT_FIND_EQ, 1}, 
	{SEPARATOR, 0}, {BABIT_FIND_EQ, 0}, {SEPARATOR, 0},
        {BABIT_FIND_EQ, -1}, {BABIT_FIND_GT, -2},
	{BABIT_FIND_EQ, -2}, {BABIT_FIND_GT, -3},
	{BABIT_FIND_EQ, -3},
	{BABIT_FIND_GT, -4}
};
#define NTestCodes (sizeof (TestCodes)/sizeof (TestCodes[0]))

static void FindExTestSub2 (const BABIT_ROOT *rootp, BABIT_KEY *keyp)
{
    Block *obj;
    int i;

/*  printf ("  Around %6d:", key); */
    for (i = 0; i < NTestCodes; ++i) {
        if (TestCodes[i].option == SEPARATOR) {
	    printf (" |");
	} else {
 	    obj = BabitFindEx (rootp, keyp, TestCodes[i].option,
                               TestCodes[i].value, &bcb);
	    if (obj==NULL) printf ("  ***");
	    else           printf (" %4d", obj->value);
	}
    }
    printf ("\n");
}

static void FindExTestTitle (void)
{
    int i;
    int option, value;
    char tmp [16];
    char *p;
    
    for (i=0; i<NTestCodes; ++i) {
        option = TestCodes[i].option;
	value = TestCodes[i].value;
	if (option == SEPARATOR) {
	    printf (" |");
	    continue;
	}
	switch (option) {
	case BABIT_FIND_LT: p =  "<"; break;
	case BABIT_FIND_LE: p = "<="; break;
	case BABIT_FIND_EQ: p = "=="; break;
	case BABIT_FIND_GE: p = ">="; break;
	case BABIT_FIND_GT: p =  ">"; break;
	default: p = "?"; break;
	}
	sprintf (tmp, "%s%d", p, value);
	printf (" %4s",  tmp);
    }
    printf ("\n");
    for (i=0; i<NTestCodes; ++i) {
        option = TestCodes[i].option;
	value = TestCodes[i].value;
	if (option == SEPARATOR)
	    printf ("-+");
	else
	    printf ("-----");
    }
    printf ("\n");
}

static void FindExTestSub (BABIT_ROOT *rootp, BABIT_KEY *keyp)
{
    Block *obj;

/*  printf ("         Testing %d\n", key); */
    obj = BabitFind (rootp, keyp, &bcb);
    FindExTestSub2 (rootp, keyp);
    if (debug) {
        if (obj) {
	    BabitDelete (rootp, obj, &bcb);
	    FindExTestSub2 (rootp, keyp);
	    BabitInsert (rootp, obj, &bcb);
        }
    }
}

#define EXTESTS 4

static void FindExTest (BABIT_ROOT *rootp, int key)
{
    int i, j;
    int kf;

    kf = key/EXTESTS;
    printf ("Testing FindExTest: (N: %d, test-values:", key);
    for (i=0; i<=EXTESTS; ++i) {
        j=i*kf;
        printf (" %d", j);
    }
    printf (")\n");
    
    FindExTestTitle ();
    for (i=0; i<=EXTESTS; ++i) {
        j=i*kf;
        FindExTestSub (rootp, &j);
    }
}

static void WalkTestSub (const BABIT_ROOT *root, BabitWalkStruct *pbw, 
                         const char *title)
{
    Block *b;
    int d;
    
    printf ("%s: ", title);
    d = 0;
    pbw->first = 1;
    while ((b = BabitWalk (root, pbw, &bcb)) != NULL) {
        if (d==pbw->depth)
	    putchar (' ');
	else if (d<pbw->depth) {
	    putchar (' ');
            while (d<pbw->depth) { putchar ('('); ++d; }
	} else if (d>pbw->depth) {
	    while (d>pbw->depth) { putchar (')'); --d; }
	    putchar (' ');
	}    
        printf ("%d", b->value);
    }
    while (d>0) { putchar (')'); --d; }
    printf ("\n");
}

static void WalkTest (const BABIT_ROOT *root, int option)
{
    BabitWalkStruct bw;

    printf ("Testing BabitWalk\n");
    if (option&1) {
    bw.options = BABIT_WALK_PREORDER;
    WalkTestSub (root, &bw, "  preorder");
    }

    if (option&2) {
    bw.options = BABIT_WALK_INORDER;
    WalkTestSub (root, &bw, "  inorder");
    }

    if (option&4) {
    bw.options = BABIT_WALK_POSTORDER;
    WalkTestSub (root, &bw, "  postorder");
    }

    if (option&8) {
    bw.options = BABIT_WALK_ALLORDER;
    WalkTestSub (root, &bw, "  allorder");
    }
}

static void DeleteTest (BABIT_ROOT *rootp, int option)
{
    int i, n, dom, he;
    Block *obj;

    if (check) BabitCheck (rootp, &dom, &he, &bcb);
/*  printf ("Before delete height=%d, dom=%d\n", he, dom); */

    n = 0;
    while (*rootp) {
/*	i = root->value;*/
        obj = BabitGetMin (rootp, option, &bcb);
	i = obj->value;
	if (debug) printf ("Most torlom a %d-t\n", i);
/*	FindExTest (rootp, i);*/
	if (BabitDelete (rootp, obj, &bcb)) {
	    fprintf (stderr, "*** Erase error on %d\n", i);
	}
/*	FindExTest (rootp, i);*/
/*	WalkTest (rootp, 1);*/
	if (check) BabitCheck (rootp, &dom, &he, &bcb);
/*	printf ("After deleting %d height=%d, dom=%d\n", i, he, dom);*/
	free (obj);
	++n;
    }
    printf ("Torolve %d tetel\n", n);
}

static void InsertTest (BABIT_ROOT *root, int *vect, int vectn)
{
    int i, n, newn, dom, he;
    Block *obj, *dupl;

    BabitCheck (root, &n, &he, &bcb);

    newn = 0;
    for (i=0;i < vectn; ++i) {
	obj = malloc (sizeof (Block));
	if (obj == NULL) {
	    fprintf (stderr, "*** Out of memory\n");
	    exit (99);
	}
	obj->value = vect[i];

	if (debug) printf ("Insert: Most jon a %d\n", vect[i]);
	dupl = BabitInsert (root, obj, &bcb);
	if (dupl) {
	    fprintf (stderr, "*** %d is duplicated\n", vect[i]);
	    exit (100);
	}
	++n;
        ++newn;
	if (check) {
	    BabitCheck (root, &dom, &he, &bcb);
	    if (dom != n) {
	        fprintf (stderr, "Elemek szama=%d <> %d\n", dom, n);
	    }
/*	printf ("After %d height=%d, dom=%d\n", i, he, dom); */
        }
    }
    printf ("Felvittem %d tetelt (osszes=%d)\n", newn, n);
}

static void GenSeq (int *vect, int n, int method)
{
    int i, r, tmp;
    int max;

    max = n-1;
    if (method > 0) {
        for (i=0; i<=max; ++i)
            vect[i] = i;
    } else {
        for (i=0; i<=max; ++i)
            vect[i] = max-i;
    }

    if (method == 0) {
        for (i=max; i>0; --i) {
            r = rand () % (i+1);
            tmp = vect[i];
            vect[i] = vect[r];
            vect[r] = tmp;
        }
    }
}
